local MAJOR = "StarLibBuffer-1.0"
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibBuffer = LibStub:NewLibrary(MAJOR, MINOR)
if not LibBuffer then return end
local LibError = LibStub("StarLibError-1.0")
assert(LibError, MAJOR .. " requires StarLibError-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LibBuffer.__index then
	LibBuffer.__index = LibBuffer
end

-- @name StarLibBuffer:New
-- @usage StarLibBuffer:New(visitor, rows, cols, yres, xres, layers)
-- @param size The buffer's size
-- @param errorLevel The verbocity level
-- @return A new StarLibBuffer object

function LibBuffer:New(name, size, errorLevel)
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
		obj.buffer = {}
	end

	setmetatable(obj, self)

	obj:Resize(size or 0)
	obj.name = name
	obj.error = LibError:New(MAJOR, errorLevel)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	return obj	
	
end

-- @name StarLibBuffer:Del
-- @usage StarLibBuffer:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LibBuffer:Del(buff)
	if not buff then buff = self end
	pool[buff] = true
	for i, v in ipairs(objects) do
		if v == buff then
			objectsDict[buff] = nil
			tremove(objects, i)
		end
	end
	buff:Clear()
	buff.error:Del()
end

function LibBuffer:Resize(size)
	if self:Size() < size then
		for i = self:Size(), size - 1 do
			self.buffer[i] = ' '
		end
	end
	if self:Size() > size then
		for i = self:Size(), size - 1, -1 do
			self.buffer[i] = nil
		end
	end
	self.size = size
end

function LibBuffer:Replace(pos, val)
	if pos >= self.size then return end
	if type(val) ~= "string" or val == "" then
		val = ' '
	end
	self.buffer[pos] = val:sub(1, 1)
end

function LibBuffer:Clear()
	table.wipe(self.buffer)
	for i = 0, self.size - 1 do
		self.buffer[i] = ' '
	end
end

function LibBuffer:AsString()
	local str = ""
	for i = 0, self:Size() - 1 do
		if type(self.buffer[i]) == "string" then
			str = str .. self.buffer[i]
		else
			self.error:Print(":AsString() - Bad index: " .. i .. ", type: " .. type(self.buffer[i]) .. ", size: " .. self:Size())
			break
		end
	end
	return str
end

function LibBuffer:FromString(str)
	if type(str) ~= "string" then return end
	table.wipe(self.buffer)
	for i = 0, strlen(str) - 1 do
		self.buffer[i] = str:sub(i + 1, i + 1)
	end
	self.size = strlen(str)
end

function LibBuffer:Size()
	--[[
	local count = 0
	for k in pairs(self.buffer) do
		count = count + 1
	end
	return count
	]]
	return self.size or 0
end

function LibBuffer:Sub(num1, num2)

	if not num2 then num2 = self:Size() - 1 end
	
	local j = self:Size() - num1 + (num2 - num1)
	local t = LibBuffer:New(j)
	for i = num1, num2 do
		t.buffer[i - num1] = self.buffer[i]
	end
	return t
end

function LibBuffer:Copy(other)
	for k, v in pairs(self.buffer) do
		self.buffer[k] = nil
	end
	for k, v in pairs(other.buffer) do
		self.buffer[k] = other.buffer[k]
	end
end

function LibBuffer:Merge(from, pos, len)
	local i = 0
	while i < len do
		self.buffer[i + pos] = from.buffer[i]
		i = i + 1
	end
end
