local MAJOR = "StarLibLCDText-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LCDText = LibStub:NewLibrary(MAJOR, MINOR)
if not LCDText then return end
local core = LibStub("StarLibCore-1.0")
assert(core, MAJOR .. " requires StarLibCore-1.0")
local LibError = LibStub("StarLibError-1.0")
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local LibBuffer = LibStub("StarLibBuffer-1.0", true)
assert(LibBuffer, MAJOR .. " requires StarLibBuffer-1.0")
local LibUtils = LibStub("StarLibUtils-1.0", true)
assert(LibUtils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

local objects = {}
local objectsDict = {}
local update
local frame = CreateFrame("Frame")

local DEFAULT_LIMIT = 50

if not LCDText.__index then
	LCDText.__index = LCDText
end

-- @name StarLCDText:New
-- @usage StarLCDText:New(visitor, rows, cols, yres, xres, layers)
-- @param visitor A StarLibCore object
-- @param rows The number of rows
-- @param cols The number of cols
-- @param yres The y resolution for characters
-- @param xres The x resolution for characters
-- @param layers The number of layers
-- @param errorLevel Error verbosity level
-- @return A new StarLCDText object

function LCDText:New(visitor, rows, cols, yres, xres, layers, errorLevel, realBlit, data)
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)
	
	tinsert(objects, obj)
	objectsDict[obj] = obj

	obj.error = LibError:New(MAJOR, errorLevel) 
	obj.visitor = visitor
	obj.LROWS = rows
	obj.LCOLS = cols
	obj.DROWS = rows
	obj.DCOLS = cols
	obj.LAYERS = layers
	obj.errorLevel = errorLevel
	obj.RealBlit = realBlit
	obj.data = data

	local n = rows * cols
	
	obj.LayoutFB = {}
	obj.TransitionFB = {}
	for i = 0, layers do
		obj.LayoutFB[i] = LibBuffer:New(MAJOR .. ".LayoutFB", n, self.errorLEvel)
		obj.TransitionFB[i] = LibBuffer:New(MAJOR .. ".TransitionFB", n, self.errorLevel)
	end
	obj.DisplayFB = LibBuffer:New(MAJOR .. ".DisplayFB", n, self.errorLevel)
	
	return obj	
	
end

-- @name StarLCDText:Del
-- @usage StarLCDText:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LCDText:Del(lcd)
	if not lcd then lcd = self end
	
	if type(lcd) == "table" then
		pool[lcd] = true
		for i, v in ipairs(objects) do
			if v == lcd then
				objectsDict[lcd] = nil
				tremove(objects, i)
			end
		end
		lcd.error:Del()
	end
end

function LCDText:TextBlit(row, col, height, width)
	--local lr, lc
	--local dr, dc
	--local p1, p2
	self._fb = LibBuffer:New(MAJOR .. "._fb", self.LROWS * self.LCOLS)
	
	self._r = row
	--[[
	while r < self.LROWS and r < row + height do
		local c = col
		while c < self.LCOLS and c < col + width do
			local n = r * self.LCOLS + c
			local l = 0
			while l < self.LAYERS do
				if self.LayoutFB[l].buffer[n] ~= ' ' then
					fb:Replace(n, self.LayoutFB[l].buffer[n])
					
				end
				l = l + 1
			end
			c = c + 1
		end
		r = r + 1
	end
	]]
	self._fb:Copy(self.LayoutFB[0])

	--/* loop over layout rows */
	self._lr = row
	
	while self._lr < self.LROWS and self._lr < row + height do
		--/* transform layout to display row */
		self._dr = self._lr;
		--/* sanity check */
		if self._dr >= 0 and self._dr < self.DROWS then 		
			--/* loop over layout cols */
			self._lc = col
			while self._lc < self.LCOLS and self._lc < col + width do
				--/* transform layout to display column */
				self._dc = self._lc;
				--/* sanity check */
				if self._dc >= 0 and self._dc < self.DCOLS then
					--/* find start of difference */
					if self.DisplayFB.buffer[self._dr * self.DCOLS + self._dc] ~= self._fb.buffer[self._lr * self.LCOLS + self._lc] then
					--if string.sub(self.DisplayFB, dr * self.DCOLS + dc, 1) ~= string.sub(fb, lr * self.LCOLS + lc) then
						--/* find end of difference */
						self._p1 = self._dc
						self._p2 = self._p1
						while self._lc < self.LCOLS and self._lc < col + width do
							--/* transform layout to display column */
							self._dc = self._lc;
							--/* sanity check */
							if self._dc >= 0 and self._dc < self.DCOLS then
								if self.DisplayFB.buffer[self._dr * self.DCOLS + self._dc] ~= self._fb.buffer[self._lr * self.LCOLS + self._lc] then
									self._p2 = self._dc;
								end
							end
							self._lc  = self._lc + 1
						end

						--/* send to display */
						self._from = self._fb:Sub(self._lr * self.LCOLS + self._p1)
						self.DisplayFB:Merge(self._from, self._dr * self.DCOLS + self._p1, self._p2 - self._p1 + 1)
						if (self.RealBlit) then
							self._tmp = self.DisplayFB:Sub(self._dr * self.DCOLS + self._p1)
							self:RealBlit(self.data, self._dr, self._p1, self._tmp.buffer, self._p2 - self._p1 + 1)
							self._tmp:Del()
							self._tmp = nil
						end
						self._from:Del()
					end
				end
				self._lc = self._lc + 1
			end
		end
		self._lr = self._lr + 1
	end
	self._fb:Del()
end

function LCDText.TextDraw(widget, data)
	widget._lcd = widget.visitor.lcd
	widget._n  = widget.widget.row * widget._lcd.LCOLS + widget.widget.col
	widget._len = strlen(widget.buffer)
	widget._num = 1
	for i = widget._n, widget._n + widget._len do
		if i / widget._lcd.LROWS >= widget._lcd.LCOLS then
			break
		end
		widget._lcd.LayoutFB[widget.widget.layer]:Replace(i, widget.buffer:sub(widget._num, widget._num))
		widget._num = widget._num + 1
	end

	widget._lcd:TextBlit(widget.widget.row, widget.widget.col, 1, widget._len)
end